/*////////////////////////////////////////////////////////////////////////
//																		//
//  Title: cbc_table.h - Node's beacon table class define file			//
//																		//
//																		//
//  Function: Defines all classes responsible for manage				//
//	      the received becons on each node								//
//																		//
//																		//
//  Author: Rodolfo Oliveira - June 2004								//
//	    rado@uninova.pt													//
//	Modified by: Miguel Luís - September 2009							//
//		nml17338@fct.unl.pt												//
//																		//
//  New University of Lisbon / FCT - All rights reserved				//
//																		//
////////////////////////////////////////////////////////////////////////*/
#ifndef bc_table_h
#define bc_table_h

#include <config.h>
#include <timer-handler.h>
#include "beaconing_parameters.h"

#include <vector>
#include <map>

typedef std::vector<u_int32_t> vector_lst;

class Beacons_tab_node;
class Beacons_tab;

//------------------------------------------------------------------------
// Node of address list
class Addr_list_node
{
public:
	Addr_list_node();

	nsaddr_t addr;  //
	Addr_list_node* next;
};

//------------------------------------------------------------------------
// Address list
class Addr_list
{
public:
	Addr_list();
	~Addr_list();
	void Add_elem(nsaddr_t *address);
	int Get_num_elems();
	int Get_elem_at(int pos, nsaddr_t *result);
	int Delete_all();

	// to charge the table from a beacon message
	void Charge_list(char* ptr, u_int8_t bc_gl_lis_elems);

protected:
	Addr_list_node* m_head;
	Addr_list_node* m_tail;
	int m_elems;
};


//------------------------------------------------------------------------
// TIMER to timeout(delete) the beacon table entry
class Timer_delete : public TimerHandler {
private:
	virtual void expire(Event *e);
    Beacons_tab_node *a_;

public:
    Timer_delete(Beacons_tab_node* a) : TimerHandler() {
    	a_=a;
    	double interval = BEACON_TIMEOUT;
    	this->resched(interval);
   	};
};

//------------------------------------------------------------------------
// TIMER to increment assoctiativity 
class Timer_bPeriod : public TimerHandler {	
private:
	virtual void expire(Event *e);
	Beacons_tab_node *a_;
public:
    Timer_bPeriod(Beacons_tab_node* a) : TimerHandler() {
    	a_=a;
   	};
};

//------------------------------------------------------------------------
// Node's list of received beacons addresses
class Beacons_tab_node
{
friend class Timer_delete;
friend class Timer_bPeriod;

public:
	Beacons_tab_node();
	~Beacons_tab_node();
 
	nsaddr_t   bc_elem_node; 			// neighbor address
	u_int32_t  bc_elem_seq_num;
	u_int16_t  bc_elem_associativity;
	u_int16_t  bc_elem_bl_associativity;
	nsaddr_t   bc_elem_gl_addr;
	Addr_list  *bc_elem_gl_list;
  
	Timer_delete timer;
	Timer_bPeriod count_beaconsTimer; 
  
	Beacons_tab_node *next;
	Beacons_tab *tab;

};


//------------------------------------------------------------------------
// List of received beacons addresses
class Beacons_tab
{
public:
	Beacons_tab();
	~Beacons_tab();
 
	void Add_beacon(u_int32_t bc_seq_num, u_int32_t bc_associativity, nsaddr_t bc_gl_addr, nsaddr_t *received_from);
	
	u_int16_t Return_assoc_of_node(nsaddr_t received_from);
	std::map<nsaddr_t, u_int16_t> Return_assoc_of_all_nodes();
 
	void Delete_beacon(nsaddr_t *received_from);
	int Get_num_elems();
	void Print();

	Beacons_tab_node* Get_elem_at(int pos);
	nsaddr_t address; 				//address of this node

	Beacons_tab_node* m_head;
	Beacons_tab_node* m_tail;

	void Refresh_group_leader();

protected:
	bool Refresh_stability(int *ro_max, int *v_max);
	void Refresh_is_BGL();

	int m_elems;

public:
	bool m_is_stable; 				// this node is stable ??
	bool m_is_bgl;
	
	// Information to build send beacon procedure
	u_int16_t  bc_associativity;
	nsaddr_t   bc_gl_addr;			// -1 ==> no group leader
	u_int8_t   bc_gl_lis_elems;		// num of elements contained in bc_gl_list - list of group leaders in the neigborhood
	char*      bc_gl_list;

	void insert_in_sort_list(u_int32_t ask,std::vector<u_int32_t> &vectr);
	
public:

	// Variables for link duration and pdf statistics
	static int link_counter; 		// Total links ocurred 
	static vector_lst histogram; 	// Histogram with number of cleared links 
	void print_link_duration_stats();
  
};
#endif //bc_table_h
